import json
import os
from typing import Dict, TextIO, Protocol, Union, Iterable, Optional
import logging
from lusid_workflow.extensions.proxy_config import ProxyConfig
from lusid_workflow.extensions.api_configuration import ApiConfiguration
from lusid_workflow.extensions.file_access_token import FileAccessToken

logger = logging.getLogger(__name__)

ENVIRONMENT_CONFIG_KEYS = {
    "token_url": "FBN_TOKEN_URL",
    "api_url": "FBN_WORKFLOW_URL",
    "previous_api_url": "FBN_LUSID_WORKFLOW_API_URL",
    "username": "FBN_USERNAME",
    "password": "FBN_PASSWORD",
    "client_id": "FBN_CLIENT_ID",
    "client_secret": "FBN_CLIENT_SECRET",
    "app_name": "FBN_APP_NAME",
    "certificate_filename": "FBN_CLIENT_CERTIFICATE",
    "proxy_address": "FBN_PROXY_ADDRESS",
    "proxy_username": "FBN_PROXY_USERNAME",
    "proxy_password": "FBN_PROXY_PASSWORD",
    "access_token": "FBN_ACCESS_TOKEN"
}

SECRETS_FILE_CONFIG_KEYS = {
    "token_url": "tokenUrl",
    "api_url": "workflowUrl",
    "previous_api_url": "lusid_workflowUrl",
    "username": "username",
    "password": "password",
    "client_id": "clientId",
    "client_secret": "clientSecret",
    "app_name": "applicationName",
    "certificate_filename": "clientCertificate",
    "proxy_address": "address",
    "proxy_username": "username",
    "proxy_password": "password",
    "access_token": "accessToken"
}


class ConfigurationLoader(Protocol):
    """
    The ApiConfigurationLoader is responsible for populating the API and Proxy configuration from a secrets file or
    environment variables with preference given to the secrets file.
    """

    def load_config(self) -> Dict[str, str]:
        pass


class SecretsFileConfigurationLoader:
    """Configuration Loader for reading secrets from a json secrets file
    """
    def __init__(
        self,
        api_secrets_file: Union[TextIO, str]
    ):
        """Create SecretsFileConfigurationLoader

        Parameters
        ----------
        api_secrets_file : Union[TextIO, str]
            File to load secrets from
        """
        self._api_secrets_file = api_secrets_file or ""

    def load_config(self) -> Dict[str, str]:
        """reads config from the provided secrets file

        Returns
        -------
        Dict[str, str]
            dictionary that can be loaded into an ApiConfiguration object
        """
        # The secrets file is a nested dictionary, set the names of the top level keys
        logger.debug(f"loading config from secrets file: {self._api_secrets_file}")
        api_config_key = "api"
        proxy_config_key = "proxy"
        try:
            try:
                config = json.load(self._api_secrets_file)
            except AttributeError:
                with open(self._api_secrets_file) as api_secrets_file:
                    config = json.load(api_secrets_file)
        except OSError:
            logger.warning(f"Unable to open secrets file {self._api_secrets_file}")
            return {}
        except json.JSONDecodeError:
            logger.warning("unable to deserialise contents of secrets file to json")
            return {}

        api_config_section = config.get(api_config_key, {})
        populated_api_config_values = {
            key: api_config_section.get(value)
            for key, value in SECRETS_FILE_CONFIG_KEYS.items()
            if "proxy" not in key
        }
        if not populated_api_config_values["api_url"]:
            populated_api_config_values["api_url"] = populated_api_config_values["previous_api_url"]
        del(populated_api_config_values["previous_api_url"])
        proxy_config_section = config.get(proxy_config_key, {})
        populated_proxy_values = {
            key: proxy_config_section.get(value)
            for key, value in SECRETS_FILE_CONFIG_KEYS.items()
            if "proxy" in key
        }
        populated_config_dict = {
            **populated_api_config_values,
            **populated_proxy_values,
        }
        return populated_config_dict


class EnvironmentVariablesConfigurationLoader:
    """ConfigurationLoader which reads config from environment variables
    """    
    def load_config(self) -> Dict[str, str]:
        """reads config from environment variables

        Returns
        -------
        Dict[str, str]
            dictionary that can be loaded into an ApiConfiguration object
        """
        logger.debug("loading config from environment variables")

        populated_api_config_values = {
            key: os.environ.get(value)
            for key, value in ENVIRONMENT_CONFIG_KEYS.items()
            if "proxy" not in key
        }
        if not populated_api_config_values["api_url"]:
            populated_api_config_values["api_url"] = populated_api_config_values["previous_api_url"]
        del(populated_api_config_values["previous_api_url"])
        populated_proxy_values = {
            key: os.environ.get(value)
            for key, value in ENVIRONMENT_CONFIG_KEYS.items()
            if "proxy" in key
        }
        populated_config_dict = {
            **populated_api_config_values,
            **populated_proxy_values,
        }
        return populated_config_dict


class ArgsConfigurationLoader:
    """ConfigurationLoader which loads in config from kwargs in constructor
    """
    def __init__(self,
        token_url:Optional[str]=None,
        api_url:Optional[str]=None,
        username:Optional[str]=None,
        password:Optional[str]=None,
        client_id:Optional[str]=None,
        client_secret:Optional[str]=None,
        app_name:Optional[str]=None,
        certificate_filename:Optional[str]=None,
        proxy_address:Optional[str]=None,
        proxy_username:Optional[str]=None,
        proxy_password:Optional[str]=None,
        access_token:Optional[str]=None
        ):
        """kwargs passed to this constructor used to build ApiConfiguration
        """
        self.__token_url = token_url
        self.__api_url = api_url
        self.__username = username
        self.__password = password
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.__app_name = app_name
        self.__certificate_filename = certificate_filename
        self.__proxy_address = proxy_address
        self.__proxy_username = proxy_username
        self.__proxy_password = proxy_password
        self.__access_token = access_token

    def load_config(self) -> Dict[str, str]:
        """load configuration from kwargs passed to constructor

        Returns
        -------
        Dict[str, str]
            dictionary that can be loaded into an ApiConfiguration object
        """
        logger.debug("loading config from arguments passed to ArgsConfigurationLoader")
        return {
            "token_url" : self.__token_url, 
            "api_url" : self.__api_url, 
            "username" : self.__username, 
            "password" : self.__password, 
            "client_id" : self.__client_id, 
            "client_secret" : self.__client_secret, 
            "app_name" : self.__app_name, 
            "certificate_filename" : self.__certificate_filename, 
            "proxy_address" : self.__proxy_address, 
            "proxy_username" : self.__proxy_username, 
            "proxy_password" : self.__proxy_password, 
            "access_token" : self.__access_token
        }


class FileTokenConfigurationLoader:
    """ConfigurationLoader which loads in access token from file
        if FBN_ACCESS_TOKEN_FILE is set,
        or if an access_token_location is passed to the initialiser
    """

    def __init__(
        self, access_token_location: str = os.getenv("FBN_ACCESS_TOKEN_FILE", "")
    ):
        self.access_token = None
        # if neither are provided we won't want to override config from other loaders
        if access_token_location is not None and access_token_location != "":
            self.access_token = FileAccessToken(access_token_location)

    def load_config(self) -> Dict[str, Union[FileAccessToken, None]]:
        """load access token from file

        Returns
        -------
        Dict[str, str]
            dictionary that can be loaded into an ApiConfiguration object
        """
        return {"access_token": self.access_token}


default_config_loaders = (
    EnvironmentVariablesConfigurationLoader(),
    SecretsFileConfigurationLoader(api_secrets_file="secrets.json"),
    FileTokenConfigurationLoader()
)

def get_api_configuration(config_loaders: Iterable[ConfigurationLoader]) -> ApiConfiguration:
    """Read configuration from config loaders.
    Update config with values from each loader in order (last write wins).

    Parameters
    ----------
    config_loaders : Iterable[ConfigurationLoader]
        Objects that can be used to fetch config with a load_config function returning a dict.

    Returns
    -------
    ApiConfiguration
        Configuration that can be passed to an ApiClient, RefreshingToken, etc.
    """
    config = {}

    for config_loader in config_loaders:
        loaded_config = {
            key: value
            for key, value in config_loader.load_config().items()
            if value is not None and value != None
        }
        config.update(loaded_config)
    proxy_address = config.pop("proxy_address", None)
    proxy_username = config.pop("proxy_username", None)
    proxy_password = config.pop("proxy_password", None)
    # If the proxy address is missing ensure that no proxy is used in the ApiConfiguration
    if proxy_address is not None:
        config["proxy_config"] = ProxyConfig(
            address=proxy_address, username=proxy_username, password=proxy_password
        )
    else:
        config["proxy_config"] = None
    # Create and return the ApiConfiguration
    return ApiConfiguration(**config)


# coding: utf-8

"""
    FINBOURNE Workflow API

    FINBOURNE Technology  # noqa: E501

    Contact: info@finbourne.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, Dict, Optional
from pydantic.v1 import BaseModel, Field, StrictBool, StrictStr, constr

class Parameter(BaseModel):
    """
    Defines a Worker Parameter  # noqa: E501
    """
    type: constr(strict=True, min_length=1) = Field(..., description="The type of this Parameter")
    name: constr(strict=True, min_length=1) = Field(..., description="Name")
    display_name: Optional[StrictStr] = Field(None, alias="displayName", description="DisplayName")
    description: Optional[StrictStr] = Field(None, description="Description")
    required: StrictBool = Field(..., description="Required or not")
    default_value: Optional[StrictStr] = Field(None, alias="defaultValue", description="DefaultValue")
    __properties = ["type", "name", "displayName", "description", "required", "defaultValue"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Parameter:
        """Create an instance of Parameter from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # set to None if display_name (nullable) is None
        # and __fields_set__ contains the field
        if self.display_name is None and "display_name" in self.__fields_set__:
            _dict['displayName'] = None

        # set to None if description (nullable) is None
        # and __fields_set__ contains the field
        if self.description is None and "description" in self.__fields_set__:
            _dict['description'] = None

        # set to None if default_value (nullable) is None
        # and __fields_set__ contains the field
        if self.default_value is None and "default_value" in self.__fields_set__:
            _dict['defaultValue'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> Parameter:
        """Create an instance of Parameter from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return Parameter.parse_obj(obj)

        _obj = Parameter.parse_obj({
            "type": obj.get("type"),
            "name": obj.get("name"),
            "display_name": obj.get("displayName"),
            "description": obj.get("description"),
            "required": obj.get("required"),
            "default_value": obj.get("defaultValue")
        })
        return _obj



# coding: utf-8

"""
    FINBOURNE Workflow API

    FINBOURNE Technology  # noqa: E501

    Contact: info@finbourne.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from typing import Any, Dict, List, Optional
from pydantic.v1 import BaseModel, Field, StrictInt, conlist, constr, validator
from lusid_workflow.models.field_mapping import FieldMapping
from lusid_workflow.models.resource_id import ResourceId
from lusid_workflow.models.resultant_child_task_configuration import ResultantChildTaskConfiguration
from lusid_workflow.models.worker_status_triggers import WorkerStatusTriggers

class RunWorkerAction(BaseModel):
    """
    Defines a Run Worker Action  # noqa: E501
    """
    type: constr(strict=True, min_length=1) = Field(..., description="Type name for this Action")
    worker_id: ResourceId = Field(..., alias="workerId")
    worker_as_at: Optional[datetime] = Field(None, alias="workerAsAt", description="Worker AsAt")
    worker_parameters: Optional[Dict[str, FieldMapping]] = Field(None, alias="workerParameters", description="Parameters for this Worker")
    worker_status_triggers: Optional[WorkerStatusTriggers] = Field(None, alias="workerStatusTriggers")
    child_task_configurations: Optional[conlist(ResultantChildTaskConfiguration)] = Field(None, alias="childTaskConfigurations", description="Tasks can be generated from run worker results; this is the configuration")
    worker_timeout: Optional[StrictInt] = Field(None, alias="workerTimeout", description="Worker WorkerTimeout in seconds")
    __properties = ["type", "workerId", "workerAsAt", "workerParameters", "workerStatusTriggers", "childTaskConfigurations", "workerTimeout"]

    @validator('type')
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value not in ('RunWorker'):
            raise ValueError("must be one of enum values ('RunWorker')")
        return value

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> RunWorkerAction:
        """Create an instance of RunWorkerAction from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of worker_id
        if self.worker_id:
            _dict['workerId'] = self.worker_id.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each value in worker_parameters (dict)
        _field_dict = {}
        if self.worker_parameters:
            for _key in self.worker_parameters:
                if self.worker_parameters[_key]:
                    _field_dict[_key] = self.worker_parameters[_key].to_dict()
            _dict['workerParameters'] = _field_dict
        # override the default output from pydantic by calling `to_dict()` of worker_status_triggers
        if self.worker_status_triggers:
            _dict['workerStatusTriggers'] = self.worker_status_triggers.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in child_task_configurations (list)
        _items = []
        if self.child_task_configurations:
            for _item in self.child_task_configurations:
                if _item:
                    _items.append(_item.to_dict())
            _dict['childTaskConfigurations'] = _items
        # set to None if worker_as_at (nullable) is None
        # and __fields_set__ contains the field
        if self.worker_as_at is None and "worker_as_at" in self.__fields_set__:
            _dict['workerAsAt'] = None

        # set to None if worker_parameters (nullable) is None
        # and __fields_set__ contains the field
        if self.worker_parameters is None and "worker_parameters" in self.__fields_set__:
            _dict['workerParameters'] = None

        # set to None if child_task_configurations (nullable) is None
        # and __fields_set__ contains the field
        if self.child_task_configurations is None and "child_task_configurations" in self.__fields_set__:
            _dict['childTaskConfigurations'] = None

        # set to None if worker_timeout (nullable) is None
        # and __fields_set__ contains the field
        if self.worker_timeout is None and "worker_timeout" in self.__fields_set__:
            _dict['workerTimeout'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> RunWorkerAction:
        """Create an instance of RunWorkerAction from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return RunWorkerAction.parse_obj(obj)

        _obj = RunWorkerAction.parse_obj({
            "type": obj.get("type"),
            "worker_id": ResourceId.from_dict(obj.get("workerId")) if obj.get("workerId") is not None else None,
            "worker_as_at": obj.get("workerAsAt"),
            "worker_parameters": dict(
                (_k, FieldMapping.from_dict(_v))
                for _k, _v in obj.get("workerParameters").items()
            )
            if obj.get("workerParameters") is not None
            else None,
            "worker_status_triggers": WorkerStatusTriggers.from_dict(obj.get("workerStatusTriggers")) if obj.get("workerStatusTriggers") is not None else None,
            "child_task_configurations": [ResultantChildTaskConfiguration.from_dict(_item) for _item in obj.get("childTaskConfigurations")] if obj.get("childTaskConfigurations") is not None else None,
            "worker_timeout": obj.get("workerTimeout")
        })
        return _obj



# coding: utf-8

"""
    FINBOURNE Workflow API

    FINBOURNE Technology  # noqa: E501

    Contact: info@finbourne.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""


from __future__ import annotations
import pprint
import re  # noqa: F401
import json


from typing import Any, Dict, List
from pydantic.v1 import BaseModel, Field, conlist, constr

class ValueConstraints(BaseModel):
    """
    Constraints that should be applied to a Tasks fields  # noqa: E501
    """
    constraint_type: constr(strict=True, min_length=1) = Field(..., alias="constraintType", description="Whether the constraint is a suggestion or should be enforced via validation (e.g. Suggested, Validated)")
    value_source_type: constr(strict=True, min_length=1) = Field(..., alias="valueSourceType", description="The source of the acceptable values (e.g. AcceptableValues)")
    acceptable_values: conlist(Any) = Field(..., alias="acceptableValues", description="The acceptable values for the field")
    __properties = ["constraintType", "valueSourceType", "acceptableValues"]

    class Config:
        """Pydantic configuration"""
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ValueConstraints:
        """Create an instance of ValueConstraints from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ValueConstraints:
        """Create an instance of ValueConstraints from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ValueConstraints.parse_obj(obj)

        _obj = ValueConstraints.parse_obj({
            "constraint_type": obj.get("constraintType"),
            "value_source_type": obj.get("valueSourceType"),
            "acceptable_values": obj.get("acceptableValues")
        })
        return _obj
